import os
import sys
import time

from aperturedb.Connector import Connector
from aperturedb import ProgressBar

DESCRIPTOR_CLASS = "_Descriptor"
DESCRIPTOR_CONNECTION_CLASS = "_DescriptorSetToDescriptor"


class Utils(object):
    """
    **A bunch of helper methods to get information from aperturedb.**

    Args:
        object (Connector): The underlying Connector.
    """

    def __init__(self, connector: Connector, verbose=False):

        self.connector = connector.create_new_connection()
        self.verbose = verbose

    def __repr__(self):

        return self.status()

    def print(self, str):

        if self.verbose:
            print(str)

    def status(self):

        q = [{"GetStatus": {}}]

        try:
            res, blobs = self.connector.query(q)
        except:
            self.connector.print_last_response()

        return self.connector.get_last_response_str()

    def print_schema(self, refresh=False):

        self.get_schema(refresh=refresh)
        self.connector.print_last_response()

    def get_schema(self, refresh=False):

        query = [{
            "GetSchema": {
                "refresh": refresh,
            }
        }]

        res, blobs = self.connector.query(query)

        schema = {}

        try:
            schema = res[0]["GetSchema"]
        except:
            print("Cannot retrieve schema")
            self.connector.print_last_response()

        return schema

    def _create_index(self, index_type, class_name, property_key, property_type):

        q = [{
            "CreateIndex": {
                "index_type":    index_type,
                "class":         class_name,
                "property_key":  property_key,
                "property_type": property_type
            }
        }]

        try:
            res, blobs = self.connector.query(q)
            if not self.connector.last_query_ok():
                self.connector.print_last_response()
                return False
        except:
            self.connector.print_last_response()
            return False

        return True

    def _remove_index(self, index_type, class_name, property_key):

        q = [{
            "RemoveIndex": {
                "index_type":   index_type,
                "class":        class_name,
                "property_key": property_key,
            }
        }]

        try:
            res, blobs = self.connector.query(q)
            if not self.connector.last_query_ok():
                self.connector.print_last_response()
                return False
        except:
            self.connector.print_last_response()
            return False

        return True

    def create_entity_index(self, class_name, property_key, property_type):

        return self._create_index("entity", class_name,
                                  property_key, property_type)

    def create_connection_index(self, class_name, property_key, property_type):

        return self._create_index("connection", class_name,
                                  property_key, property_type)

    def remove_entity_index(self, class_name, property_key):

        return self._remove_index("entity", class_name, property_key)

    def remove_connection_index(self, class_name, property_key):

        return self._remove_index("connection", class_name, property_key)

    def count_images(self, constraints={}):

        q = [{
            "FindImage": {
                "blobs": False,
                "results": {
                    "count": True,
                }
            }
        }]

        if constraints:
            q[0]["FindImage"]["constraints"] = constraints

        try:
            res, blobs = self.connector.query(q)
            total_images = res[0]["FindImage"]["count"]
        except:
            total_images = 0
            self.connector.print_last_response()

        return total_images

    def count_bboxes(self, constraints={}):

        q = [{
            "FindBoundingBox": {
                "blobs": False,
                "results": {
                    "count": True,
                }
            }
        }]

        if constraints:
            q[0]["FindBoundingBox"]["constraints"] = constraints

        try:
            res, blobs = self.connector.query(q)
            total_connections = res[0]["FindBoundingBox"]["count"]
        except:
            total_connections = 0
            self.connector.print_last_response()

        return total_connections

    def count_entities(self, entity_class, constraints={}):

        q = [{
            "FindEntity": {
                "with_class": entity_class,
                "results": {
                    "count": True,
                }
            }
        }]

        if constraints:
            q[0]["FindEntity"]["constraints"] = constraints

        try:
            res, blobs = self.connector.query(q)
            total_entities = res[0]["FindEntity"]["count"]
        except:
            total_entities = 0
            self.connector.print_last_response()

        return total_entities

    def count_connections(self, connections_class, constraints={}):

        q = [{
            "FindConnection": {
                "with_class": connections_class,
                "results": {
                    "count": True,
                }
            }
        }]

        if constraints:
            q[0]["FindConnection"]["constraints"] = constraints

        try:
            res, blobs = self.connector.query(q)
            total_connections = res[0]["FindConnection"]["count"]
        except:
            total_connections = 0
            self.connector.print_last_response()

        return total_connections

    def add_descriptorset(self, name, dim, metric="L2", engine="FaissFlat"):

        query = [{
            "AddDescriptorSet": {
                "name":       name,
                "dimensions": dim,
                "metric":     metric,
                "engine":     engine
            }
        }]

        response, arr = self.connector.query(query)

        expected = [{
            "AddDescriptorSet": {
                "status": 0,
            }
        }]

        if response != expected:
            print("Error inserting set", name)
            self.connector.print_last_response()

    def count_descriptorsets(self):

        q = [{
            "FindDescriptorSet": {
                "results": {
                    "count": True,
                }
            }
        }]

        try:
            res, blobs = self.connector.query(q)
            total_descriptor_sets = res[0]["FindDescriptorSet"]["count"]
        except:
            total_descriptor_sets = 0
            self.connector.print_last_response()

        return total_descriptor_sets

    def get_descriptorset_list(self):

        q = [{
            "FindDescriptorSet": {
                "results": {
                    "all_properties": True,
                }

            }
        }]

        sets = []
        try:
            res, _ = self.connector.query(q)

            sets = [ent["_name"]
                    for ent in res[0]["FindDescriptorSet"]["entities"]]
        except:
            self.connector.print_last_response()

        return sets

    def remove_descriptorset(self, set_name):

        q = [{
            "FindDescriptorSet": {
                "_ref": 1,
                "with_name": set_name,

            }
        }, {
            "DeleteDescriptorSet": {
                "ref": 1
            }
        }]

        try:
            res, _ = self.connector.query(q)
            if not self.connector.last_query_ok():
                self.connector.print_last_response()
                return False
        except:
            self.connector.print_last_response()
            return False

        return True

    def _remove_objects(self, type, class_name, batch_size):

        if type == "entities":
            cmd = "Entity"
            count = self.count_entities(class_name)
        elif type == "connections":
            cmd = "Connection"
            count = self.count_connections(class_name)
        else:
            raise ValueError("Type must be either 'entities' or 'connections'")

        total = count

        pb = ProgressBar.ProgressBar()

        find = "Find" + cmd
        dele = "Delete" + cmd
        while count > 0:

            q = [{
                find: {
                    "_ref": 1,
                    "with_class": class_name,
                    "results": {
                        "limit": batch_size
                    }
                }
            }, {
                dele: {
                    "ref": 1
                }
            }]

            res, _ = self.connector.query(q)

            if not self.connector.last_query_ok():
                self.connector.print_last_response()
                return False

            count -= batch_size

            if self.verbose:
                pb.update(abs((count - total) / total))

        return True

    def remove_entities(self, class_name, batch_size=1000):

        return self._remove_objects("entities", class_name, batch_size)

    def remove_connections(self, class_name, batch_size=1000):

        return self._remove_objects("connections", class_name, batch_size)

    def remove_all_descriptorsets(self):

        self.print("Removing indexes...")

        idx_props = self.get_indexed_props(DESCRIPTOR_CLASS)

        for idx in idx_props:
            self.remove_entity_index(DESCRIPTOR_CLASS, idx)

        self.print("Done removing indexes.")

        self.print("Removing connections...")
        if not self.remove_connections(DESCRIPTOR_CONNECTION_CLASS):
            print("Error removing connections.")
            return False
        self.print("Done removing connections.")

        self.print("Removing descriptors...")
        if not self.remove_entities(DESCRIPTOR_CLASS):
            print("Error removing descriptors.")
            return False
        self.print("Done removing descriptors.")

        sets = self.get_descriptorset_list()

        self.print("Removing sets...")
        for s in sets:
            self.print("Removing {}...".format(s))
            self.remove_descriptorset(s)

        self.print("Done removing sets.")

    def get_indexed_props(self, class_name, type="entities"):
        """
        Returns all the indexed properties for a given class.
        """

        if type is not "entities" and type is not "connections":
            raise ValueError("Type must be either 'entities' or 'connections'")

        # TODO we should probably set refresh=True so we always
        # check the latest state, but it may take a long time to complete.
        schema = self.get_schema(refresh=False)

        indexed_props = schema[type]["classes"][class_name]["properties"]
        indexed_props = [
            k for k in indexed_props.keys() if indexed_props[k][1]]

        return indexed_props



import time
import unittest
import os

import torch
import torch.distributed as dist

from test_Base import TestBase

from aperturedb import Connector, Utils
from aperturedb import Images
from aperturedb import PyTorchDataset


class TestTorchDatasets(TestBase):

    '''
        These tests need to be run after the Loaders, because it uses
        data inserted by the loaders.
        TODO: We should change this so that we do not depend on them
    '''

    def test_omConstraints(self):

        db = self.create_connection()

        const = Images.Constraints()
        const.greaterequal("age", 0)
        dataset = PyTorchDataset.ApertureDBDatasetConstraints(
            db, constraints=const)

        dbutils = Utils.Utils(db)
        self.assertEqual(len(dataset), dbutils.count_images())

        start = time.time()

        # Iterate over dataset.
        for img in dataset:
            if len(img[0]) < 0:
                print("Empty image?")
                self.assertEqual(True, False)

        print("\n")
        print("Throughput (imgs/s):", len(dataset) / (time.time() - start))

    def test_nativeContraints(self):

        db = self.create_connection()

        query = [{
            "FindImage": {
                "constraints": {
                    "age": [">=", 0]
                },
                "operations": [
                    {
                        "type": "resize",
                        "width": 224,
                        "height": 224
                    }
                ],
                "results": {
                    "list": ["license"]
                }
            }
        }]

        dataset = PyTorchDataset.ApertureDBDataset(
            db, query, label_prop="license")

        dbutils = Utils.Utils(db)
        self.assertEqual(len(dataset), dbutils.count_images())

        start = time.time()

        # Iterate over dataset.
        for img in dataset:
            if len(img[0]) < 0:
                print("Empty image?")
                self.assertEqual(True, False)

        print("\n")
        print("Throughput (imgs/s):", len(dataset) / (time.time() - start))

    def test_datasetWithMultiprocessing(self):

        db = self.create_connection()

        query = [{
            "FindImage": {
                "constraints": {
                    "age": [">=", 0]
                },
                "operations": [
                    {
                        "type": "resize",
                        "width": 224,
                        "height": 224
                    }
                ],
                "results": {
                    "list": ["license"]
                }
            }
        }]

        dataset = PyTorchDataset.ApertureDBDataset(
            db, query, label_prop="license")

        dbutils = Utils.Utils(db)
        self.assertEqual(len(dataset), dbutils.count_images())

        start = time.time()

        # Iterate over dataset.
        for img in dataset:
            if len(img[0]) < 0:
                print("Empty image?")
                self.assertEqual(True, False)

        print("\n")
        print("Sequential Throughput (imgs/s):",
              len(dataset) / (time.time() - start))

        # Distributed Data Loader Setup

        # Needed for init_process_group
        os.environ['MASTER_ADDR'] = 'localhost'
        os.environ['MASTER_PORT'] = '12355'

        dist.init_process_group("gloo", rank=0, world_size=1)

        # === Distributed Data Loader Sequential

        data_loader = torch.utils.data.DataLoader(
            dataset,
            batch_size=10,          # pick random values here to test
            num_workers=4,          # num_workers > 1 to test multiprocessing works
            pin_memory=True,
            drop_last=True,
        )

        start = time.time()

        # Iterate over dataset.
        for img in data_loader:
            if len(img[0]) < 0:
                print("Empty image?")
                self.assertEqual(True, False)

        print("Distributed Data Loader Sequential Throughput (imgs/s):",
              len(dataset) / (time.time() - start))

        # === Distributed Data Loader Shuffler

        # This will generate a random sampler, which will make the use
        # of batching wasteful
        sampler     = torch.utils.data.DistributedSampler(
            dataset, shuffle=True)

        data_loader = torch.utils.data.DataLoader(
            dataset,
            sampler=sampler,
            batch_size=10,          # pick random values here to test
            num_workers=4,          # num_workers > 1 to test multiprocessing works
            pin_memory=True,
            drop_last=True,
        )

        start = time.time()

        # Iterate over dataset.
        for img in data_loader:
            if len(img[0]) < 0:
                print("Empty image?")
                self.assertEqual(True, False)

        print("Distributed Data Loader Shuffle Throughput (imgs/s):",
              len(dataset) / (time.time() - start))



import time

import numpy as np

from test_Base import TestBase

from aperturedb import Utils
from aperturedb import EntityLoader, ConnectionLoader
from aperturedb import ImageLoader, BBoxLoader, BlobLoader
from aperturedb import DescriptorSetLoader, DescriptorLoader


class TestEntityLoader(TestBase):

    def test_Loader(self):

        db = self.create_connection()
        dbutils = Utils.Utils(db)

        dbutils.create_entity_index("_Image",      "id", "integer")
        dbutils.create_entity_index("_Descriptor", "id", "integer")
        dbutils.create_entity_index("Person",      "id", "integer")

        # Insert Person Nodes
        in_csv_file = "./input/persons.adb.csv"

        generator = EntityLoader.EntityGeneratorCSV(in_csv_file)

        if self.stats:
            print("\n")

        loader = EntityLoader.EntityLoader(db)
        loader.ingest(generator, batchsize=self.batchsize,
                      numthreads=self.numthreads,
                      stats=self.stats)

        self.assertEqual(len(generator), dbutils.count_entities("Person"))

        # Insert Images
        in_csv_file = "./input/images.adb.csv"

        generator = ImageLoader.ImageGeneratorCSV(
            in_csv_file, check_image=False)

        if self.stats:
            print("\n")

        loader = ImageLoader.ImageLoader(db)
        loader.ingest(generator, batchsize=self.batchsize,
                      numthreads=self.numthreads,
                      stats=self.stats)

        self.assertEqual(len(generator), dbutils.count_images())

        # Insert Connections

        in_csv_file = "./input/connections-persons-images.adb.csv"

        generator = ConnectionLoader.ConnectionGeneratorCSV(in_csv_file)

        if self.stats:
            print("\n")

        loader = ConnectionLoader.ConnectionLoader(db)
        loader.ingest(generator, batchsize=self.batchsize,
                      numthreads=self.numthreads,
                      stats=self.stats)

        self.assertEqual(
            len(generator), dbutils.count_connections("has_image"))

        # Insert BBoxes

        in_csv_file = "./input/bboxes.adb.csv"

        generator = BBoxLoader.BBoxGeneratorCSV(in_csv_file)

        if self.stats:
            print("\n")

        loader = BBoxLoader.BBoxLoader(db)
        loader.ingest(generator, batchsize=self.batchsize,
                      numthreads=self.numthreads,
                      stats=self.stats)

        self.assertEqual(len(generator), dbutils.count_bboxes())

    def test_LoaderDescriptorsImages(self):

        # Insert Images, images may be there already, and that case should
        # be handled correctly by contraints
        db = self.create_connection()

        in_csv_file = "./input/images.adb.csv"

        generator = ImageLoader.ImageGeneratorCSV(
            in_csv_file, check_image=False)

        if self.stats:
            print("\n")

        loader = ImageLoader.ImageLoader(db)
        loader.ingest(generator, batchsize=self.batchsize,
                      numthreads=self.numthreads,
                      stats=self.stats)

        dbutils = Utils.Utils(db)
        self.assertEqual(len(generator), dbutils.count_images())

        # Insert DescriptorsSet
        in_csv_file = "./input/descriptorset.adb.csv"

        generator = DescriptorSetLoader.DescriptorSetGeneratorCSV(in_csv_file)

        if self.stats:
            print("\n")

        loader = DescriptorSetLoader.DescriptorSetLoader(db)
        loader.ingest(generator, batchsize=self.batchsize,
                      numthreads=self.numthreads,
                      stats=self.stats)

        dbutils = Utils.Utils(db)
        self.assertEqual(
            len(generator), dbutils.count_entities("_DescriptorSet"))

        # Insert Descriptors

        sets = ["setA", "setB"]

        total_descriptors = 0
        for setname in sets:

            in_csv_file = "./input/" + setname + ".adb.csv"

            generator = DescriptorLoader.DescriptorGeneratorCSV(in_csv_file)

            if self.stats:
                print("\n")

            loader = DescriptorLoader.DescriptorLoader(db)
            loader.ingest(generator, batchsize=self.batchsize,
                          # TODO FIXME HELP ME DIE
                          # This tests is failing when loading in
                          # multiple threads.
                          # We set numthreads to 1 until we fix
                          # numthreads=self.numthreads,
                          numthreads=1,
                          stats=self.stats)

            dbutils = Utils.Utils(db)
            total_descriptors += len(generator)
            self.assertEqual(total_descriptors,
                             dbutils.count_entities("_Descriptor"))

    def test_BlobLoader(self):

        # Insert Person Nodes
        db = self.create_connection()

        in_csv_file = "./input/blobs.adb.csv"

        generator = BlobLoader.BlobGeneratorCSV(in_csv_file)

        if self.stats:
            print("\n")

        loader = BlobLoader.BlobLoader(db)
        loader.ingest(generator, batchsize=self.batchsize,
                      numthreads=self.numthreads,
                      stats=self.stats)

        dbutils = Utils.Utils(db)
        self.assertEqual(len(generator), dbutils.count_entities("_Blob"))

        query = [{
            "FindBlob": {
                "blobs": True,
                "results": {
                    "list": ["license"],
                    "limit": 1
                }
            }
        }]

        res, blob = db.query(query)
        self.assertEqual(len(blob), 1)

        arr = np.frombuffer(blob[0])
        self.assertEqual(arr[2], 3.3)


class TestURILoader(TestBase):

    def test_S3Loader(self):

        # Insert Images
        db = self.create_connection()
        dbutils = Utils.Utils(db)
        count_before = dbutils.count_images()

        in_csv_file = "./input/s3_images.adb.csv"
        generator = ImageLoader.ImageGeneratorCSV(
            in_csv_file, check_image=True)

        if self.stats:
            print("\n")

        loader = ImageLoader.ImageLoader(db)
        loader.ingest(generator, batchsize=self.batchsize,
                      numthreads=self.numthreads,
                      stats=self.stats)

        dbutils = Utils.Utils(db)
        self.assertEqual(len(generator), dbutils.count_images() - count_before)

    def test_HttpImageLoader(self):

        # Insert Images
        db = self.create_connection()
        dbutils = Utils.Utils(db)
        count_before = dbutils.count_images()

        in_csv_file = "./input/http_images.adb.csv"
        generator = ImageLoader.ImageGeneratorCSV(
            in_csv_file, check_image=True)

        if self.stats:
            print("\n")

        loader = ImageLoader.ImageLoader(db)
        loader.ingest(generator, batchsize=self.batchsize,
                      numthreads=self.numthreads,
                      stats=self.stats)

        dbutils = Utils.Utils(db)
        self.assertEqual(len(generator), dbutils.count_images() - count_before)



import math
import time
from threading import Thread

from aperturedb import ParallelQuery

import numpy as np


class ParallelLoader(ParallelQuery.ParallelQuery):
    """**Parallel and Batch Loader for ApertureDB**"""

    def __init__(self, db, dry_run=False):

        super().__init__(db, dry_run=dry_run)

        self.type = "element"

    def ingest(self, generator, batchsize=1, numthreads=4, stats=False):

        self.query(generator, batchsize, numthreads, stats)

    def print_stats(self):

        times = np.array(self.times_arr)
        total_queries_exec = len(times)
        inserted_elements  = self.total_actions

        print("============ ApertureDB Loader Stats ============")
        print("Total time (s):", self.total_actions_time)
        print("Total queries executed:", total_queries_exec)

        if total_queries_exec == 0:
            print("All queries failed!")

        else:
            print("Avg Query time (s):", np.mean(times))
            print("Query time std:", np.std(times))
            print("Avg Query Throughput (q/s)):",
                  1 / np.mean(times) * self.numthreads)

            msg = "(" + self.type + "/s):"
            print("Overall insertion throughput", msg,
                  self.total_actions / self.total_actions_time)

            if self.error_counter > 0:
                print("Total errors encountered:", self.error_counter)
                inserted_elements -= self.error_counter * self.batchsize
                print("Errors (%):", 100 *
                      self.error_counter / total_queries_exec)

            print("Total inserted elements:", inserted_elements)
        print("=================================================")



import ast
from aperturedb import ParallelLoader
from aperturedb import CSVParser

HEADER_NAME   = "name"
HEADER_DIM    = "dimensions"
HEADER_ENGINE = "engine"
HEADER_METRIC = "metric"
PROPERTIES    = "properties"
CONSTRAINTS   = "constraints"


class DescriptorSetGeneratorCSV(CSVParser.CSVParser):
    """**ApertureDB DescriptorSet Data loader.**

    .. note::
        Expects a csv file with the following columns:

            ``name``, ``dimensions``, ``engine``, ``metric``, ``PROP_NAME_N``, ``constraint_PROP1``

    Example csv file::

        name,dimensions,engine,metric
        dining_chairs,2048,FaissIVFFlat,L2
        chandeliers,2048,FaissIVFFlat,L2
        console_tables,2048,FaissIVFFlat,L2
        ...
    """

    def __init__(self, filename):

        super().__init__(filename)

        self.props_keys       = [x for x in self.header[4:]
                                 if not x.startswith(CSVParser.CONTRAINTS_PREFIX)]
        self.constraints_keys = [x for x in self.header[4:]
                                 if x.startswith(CSVParser.CONTRAINTS_PREFIX)]

    def __getitem__(self, idx):

        # Metrics/Engine can be of the form:
        #       "IP", or
        #       ["IP" ...]

        metrics = self.df.loc[idx, HEADER_METRIC]
        metrics = metrics if "[" not in metrics else ast.literal_eval(metrics)

        engines = self.df.loc[idx, HEADER_ENGINE]
        engines = engines if "[" not in engines else ast.literal_eval(engines)

        data = {
            "name":       self.df.loc[idx, HEADER_NAME],
            "dimensions": self.df.loc[idx, HEADER_DIM],
            "engine":     engines,
            "metric":     metrics,
        }

        properties  = self.parse_properties(self.df, idx)
        constraints = self.parse_constraints(self.df, idx)

        if properties:
            data[PROPERTIES] = properties

        if constraints:
            data[CONSTRAINTS] = constraints

        return data

    def validate(self):

        self.header = list(self.df.columns.values)

        if self.header[0] != HEADER_NAME:
            raise Exception("Error with CSV file field: " + HEADER_NAME)
        if self.header[1] != HEADER_DIM:
            raise Exception("Error with CSV file field: " + HEADER_DIM)
        if self.header[2] != HEADER_ENGINE:
            raise Exception("Error with CSV file field: " + HEADER_ENGINE)
        if self.header[3] != HEADER_METRIC:
            raise Exception("Error with CSV file field: " + HEADER_METRIC)


class DescriptorSetLoader(ParallelLoader.ParallelLoader):
    """**ApertureDB DescriptorSet Loader.**

        This class is to be used in combination with a "generator".
        The generator must be an iterable object that generated "entity_data"
        elements.

    Example::

            entity_data = {
                "class":       entity_class,
                "properties":  properties,
                "constraints": constraints,
            }
    """

    def __init__(self, db, dry_run=False):

        super().__init__(db, dry_run=dry_run)

        self.type = "descriptorSet"

    def generate_batch(self, entity_data):

        q = []

        for data in entity_data:

            ae = {
                "AddDescriptorSet": data
            }

            if PROPERTIES in data:
                ae["AddDescriptorSet"][PROPERTIES] = data[PROPERTIES]

            if CONSTRAINTS in data:
                ae["AddDescriptorSet"]["if_not_found"] = data[CONSTRAINTS]

            q.append(ae)

        if self.dry_run:
            print(q)

        return q, []



from aperturedb import ParallelLoader
from aperturedb import CSVParser

CONNECTION_CLASS = "ConnectionClass"
PROPERTIES  = "properties"
CONSTRAINTS = "constraints"


class ConnectionGeneratorCSV(CSVParser.CSVParser):
    """**ApertureDB Connection Data loader.**

    .. note::
        Expects a csv file with the following columns:

            ``ConnectionClass``, ``Class1@PROP_NAME``, ``Class2@PROP_NAME`` ... ``PROP_NAME_N``, ``constraint_PROP1``

    Example csv file::

        ConnectionClass,_Image@id,Person@id,confidence,id,constaint_id
        has_image,321423532,42342522,0.4354,5432543254,5432543254
        has_image,321423532,53241521,0.6432,98476542,98476542
        ...
    """

    def __init__(self, filename):

        super().__init__(filename)

        self.props_keys       = [x for x in self.header[3:]
                                 if not x.startswith(CSVParser.CONTRAINTS_PREFIX)]

        self.constraints_keys = [x for x in self.header[3:]
                                 if x.startswith(CSVParser.CONTRAINTS_PREFIX)]

        self.ref1_class   = self.header[1].split("@")[0]
        self.ref1_key     = self.header[1].split("@")[1]
        self.ref2_class   = self.header[2].split("@")[0]
        self.ref2_key     = self.header[2].split("@")[1]

    def __getitem__(self, idx):

        data = {
            "class":      self.df.loc[idx, CONNECTION_CLASS],
            "ref1_class": self.ref1_class,
            "ref1_key":   self.ref1_key,
            "ref1_val":   self.df.loc[idx, self.header[1]],
            "ref2_class": self.ref2_class,
            "ref2_key":   self.ref2_key,
            "ref2_val":   self.df.loc[idx, self.header[2]],
        }

        properties  = self.parse_properties(self.df, idx)
        constraints = self.parse_constraints(self.df, idx)

        if properties:
            data[PROPERTIES] = properties

        if constraints:
            data[CONSTRAINTS] = constraints

        return data

    def validate(self):

        self.header = list(self.df.columns.values)

        if self.header[0] != CONNECTION_CLASS:
            raise Exception(
                "Error with CSV file field: 0 - " + CONNECTION_CLASS)
        if "@" not in self.header[1]:
            raise Exception("Error with CSV file field: 1")
        if "@" not in self.header[2]:
            raise Exception("Error with CSV file field: 2")


class ConnectionLoader(ParallelLoader.ParallelLoader):
    """**ApertureDB Connection Loader.**

        This class is to be used in combination with a "generator".
        The generator must be an iterable object that generated "entity_data"
        elements.

    Example::

            entity_data = {
                "class":       connection_class,
                "properties":  properties,
                "constraints": constraints,
            }
    """

    def __init__(self, db, dry_run=False):

        super().__init__(db, dry_run=dry_run)

        self.type = "connection"

    def generate_batch(self, entity_data):

        q = []

        ref_counter = 1

        for data in entity_data:

            try:

                ref_src = ref_counter
                ref_counter += 1
                fe_a = {
                    "FindEntity": {
                        "_ref": ref_src,
                        "with_class": data["ref1_class"],
                    }
                }

                fe_a["FindEntity"]["constraints"] = {}
                fe_a["FindEntity"]["constraints"][data["ref1_key"]] = [
                    "==", data["ref1_val"]]
                q.append(fe_a)

                ref_dst = ref_counter
                ref_counter += 1
                fe_b = {
                    "FindEntity": {
                        "_ref": ref_dst,
                        "with_class": data["ref2_class"],
                    }
                }

                fe_b["FindEntity"]["constraints"] = {}
                fe_b["FindEntity"]["constraints"][data["ref2_key"]] = [
                    "==", data["ref2_val"]]
                q.append(fe_b)

                ae = {
                    "AddConnection": {
                        "class": data["class"],
                        "src": ref_src,
                        "dst": ref_dst,
                    }
                }

                if PROPERTIES in data:
                    ae["AddConnection"][PROPERTIES] = data[PROPERTIES]

                if CONSTRAINTS in data:
                    ae["AddConnection"]["if_not_found"] = data[CONSTRAINTS]

                q.append(ae)
            except KeyError as ex:
                error_str = "ERROR: ConnectionLoader::generate_batch():" + \
                    " expected key='" + ex.args[0] + "' in '" + str(data) + \
                    "'. Ignored."
                print(error_str)
                sys.stdout.write(error_str + "\n")

        if self.dry_run:
            print(q)

        return q, []



import time
import requests
import os
from os import path

import cv2
import numpy as np

from aperturedb import ParallelLoader
from aperturedb import CSVParser
from aperturedb import ProgressBar

HEADER_PATH = "filename"
HEADER_URL  = "url"


class VideoDownloaderCSV(CSVParser.CSVParser):
    """
        **ApertureDB Video Downloader.**

    .. important::

        Expects a csv file with AT LEAST a ``url`` column, and
        optionally a ``filename`` field.
        If ``filename`` is not present, it is taken from the url.
    """

    def __init__(self, filename, check_video=True):

        self.has_filename = False
        self.check_video = check_video

        super().__init__(filename)

    def __getitem__(self, idx):

        url = self.df.loc[idx, HEADER_URL]

        if self.has_filename:
            filename = self.df.loc[idx, HEADER_PATH]
        else:
            filename = self.url_to_filename(url)

        return url, filename

    def url_to_filename(self, url):

        filename = url.split("/")[-1]
        folder = "/tmp/videos/"

        return folder + filename

    def validate(self):

        self.header = list(self.df.columns.values)

        if HEADER_URL not in self.header:
            raise Exception("Error with CSV file field: url. Must be a field")

        if HEADER_PATH in self.header:
            self.has_filename = True


class VideoDownloader(ParallelLoader.ParallelLoader):

    def __init__(self, db, dry_run=False):

        super().__init__(db, dry_run=dry_run)

        self.type = "video"

        self.check_video = False

    def check_if_video_is_ok(self, filename, url):

        if not os.path.exists(filename):
            return False

        try:
            a = cv2.VideoCapture(filename)
            if a.isOpened() == False:
                print("Video present but error reading it:", url)
                return False
        except:
            print("Video present but error decoding:", url)
            return False

        return True

    def download_video(self, url, filename):

        start = time.time()

        if self.check_video and self.check_if_video_is_ok(filename, url):
            return

        folder = os.path.dirname(filename)
        if not os.path.exists(folder):
            os.makedirs(folder, exist_ok=True)

        videodata = requests.get(url)
        if videodata.ok:
            fd = open(filename, "wb")
            fd.write(videodata.content)
            fd.close()

            try:
                a = cv2.VideoCapture(filename)
                if a.isOpened() == False:
                    print("Downloaded Video size error:", url)
                    os.remove(filename)
                    self.error_counter += 1
            except:
                print("Downloaded Video cannot be decoded:", url)
                os.remove(filename)
                self.error_counter += 1
        else:
            print("URL not found:", url)
            self.error_counter += 1

        self.times_arr.append(time.time() - start)

    def worker(self, thid, generator, start, end):

        if thid == 0 and self.stats:
            pb = ProgressBar.ProgressBar("download_progress.txt")

        for i in range(start, end):

            url, filename = generator[i]

            self.download_video(url, filename)

            if thid == 0 and self.stats:
                pb.update((i - start) / (end - start))

        if thid == 0 and self.stats:
            pb.update(1)

    def print_stats(self):

        print("====== ApertureDB VideoDownloader Stats ======")

        times = np.array(self.times_arr)
        print("Avg Video download time(s):", np.mean(times))
        print("Img download time std:", np.std(times))
        print("Avg download throughput (videos/s)):",
              1 / np.mean(times) * self.numthreads)

        print("Total time(s):", self.ingestion_time)
        print("Overall throughput (videos/s):",
              self.total_elements / self.ingestion_time)
        print("Total errors encountered:", self.error_counter)
        print("=============================================")



from aperturedb import ParallelLoader
from aperturedb import CSVParser

PROPERTIES  = "properties"
CONSTRAINTS = "constraints"
BLOB_PATH   = "filename"


class BlobGeneratorCSV(CSVParser.CSVParser):
    """**ApertureDB Blob Data loader.**

    .. note::
        Expects a csv file with the following columns:

            ``filename``, ``PROP_NAME_1``, ... ``PROP_NAME_N``, ``constraint_PROP1``

    Example csv file::

        filename,name,lastname,age,id,constaint_id
        /mnt/blob1,John,Salchi,69,321423532,321423532
        /mnt/blob2,Johna,Salchi,63,42342522,42342522
        ...
    """

    def __init__(self, filename):

        super().__init__(filename)

        self.props_keys       = [x for x in self.header[1:]
                                 if not x.startswith(CSVParser.CONTRAINTS_PREFIX)]
        self.props_keys       = [x for x in self.props_keys if x != BLOB_PATH]
        self.constraints_keys = [x for x in self.header[1:]
                                 if x.startswith(CSVParser.CONTRAINTS_PREFIX)]

    def __getitem__(self, idx):

        data = {}

        filename      = self.df.loc[idx, BLOB_PATH]
        blob_ok, blob = self.load_blob(filename)
        if not blob_ok:
            print("Error loading blob: " + filename)
            raise Exception("Error loading blob: " + filename)

        data["blob"] = blob

        properties  = self.parse_properties(self.df, idx)
        constraints = self.parse_constraints(self.df, idx)

        if properties:
            data[PROPERTIES] = properties

        if constraints:
            data[CONSTRAINTS] = constraints

        return data

    def load_blob(self, filename):

        try:
            fd = open(filename, "rb")
            buff = fd.read()
            fd.close()
            return True, buff
        except:
            print("BLOB ERROR:", filename)

        return False, None

    def validate(self):

        self.header = list(self.df.columns.values)

        if self.header[0] != BLOB_PATH:
            raise Exception("Error with CSV file field: " + BLOB_PATH)


class BlobLoader(ParallelLoader.ParallelLoader):
    """
    **ApertureDB Blob Loader.**

    This class is to be used in combination with a "generator".
    The generator must be an iterable object that generated "Blob_data"
    elements.

    Example::

        Blob_data = {
            "properties":  properties,
            "constraints": constraints,
        }
    """

    def __init__(self, db, dry_run=False):

        super().__init__(db, dry_run=dry_run)

        self.type = "blob"

    def generate_batch(self, Blob_data):

        q = []
        blobs = []

        for data in Blob_data:

            ae = {
                "AddBlob": {
                }
            }

            if PROPERTIES in data:
                ae["AddBlob"][PROPERTIES] = data[PROPERTIES]

            if CONSTRAINTS in data:
                ae["AddBlob"]["if_not_found"] = data[CONSTRAINTS]

            q.append(ae)

            if len(data["blob"]) == 0:
                print("WARNING: Skipping empty blob.")
                continue
            blobs.append(data["blob"])

        if self.dry_run:
            print(q)

        return q, blobs



import os
import cv2
import math
import numpy as np
from PIL import Image
from IPython.display import display
from io import BytesIO

import skimage.io as io
import matplotlib.pyplot as plt
from matplotlib.collections import PatchCollection
from matplotlib.patches import Polygon

from aperturedb import Utils


class Constraints(object):

    def __init__(self):

        self.constraints = {}

    def equal(self, key, value):

        self.constraints[key] = ["==", value]

    def greaterequal(self, key, value):

        self.constraints[key] = [">=", value]

    def greater(self, key, value):

        self.constraints[key] = [">", value]

    def lessequal(self, key, value):

        self.constraints[key] = ["<=", value]

    def less(self, key, value):

        self.constraints[key] = ["<", value]


class Operations(object):

    def __init__(self):

        self.operations_arr = []

    def get_operations_arr(self):
        return self.operations_arr

    def resize(self, width, height):

        op = {
            "type": "resize",
            "width":  width,
            "height": height,
        }

        self.operations_arr.append(op)

    def rotate(self, angle, resize=False):

        op = {
            "type": "rotate",
            "angle": angle,
            "resize": resize,
        }

        self.operations_arr.append(op)


class Images(object):

    def __init__(self, db, batch_size=100):

        self.db_connector = db

        self.images        = {}
        self.images_ids    = []
        self.images_bboxes = {}

        self.constraints = None
        self.operations  = None
        self.format      = None
        self.limit       = None

        self.search_result = None

        self.batch_size = batch_size
        self.total_cached_images = 0
        self.display_limit = 20

        self.img_id_prop     = "_uniqueid"
        self.bbox_label_prop = "_label"

    def __retrieve_batch(self, index):
        '''
        **Retrieve the batch that contains the image with the specified index**
        '''

        # Implement the query to retrieve images
        # for that batch

        total_batches = math.ceil(len(self.images_ids) / self.batch_size)
        batch_id      = int(math.floor(index / self.batch_size))

        start = batch_id * self.batch_size
        end   = min(start + self.batch_size, len(self.images_ids))

        query = []

        for idx in range(start, end):

            find = {
                "FindImage": {
                    "constraints": {
                        "_uniqueid": ["==", self.images_ids[idx]]
                    },
                }
            }

            if self.operations and len(self.operations.operations_arr) > 0:
                find["FindImage"]["operations"] = self.operations.operations_arr
            query.append(find)

        res, imgs = self.db_connector.query(query)

        if not self.db_connector.last_query_ok():
            print(self.db_connector.get_last_response_str())
            return

        for (idx, i) in zip(range(start, end), range(end - start)):

            if idx >= len(self.images_ids):
                return

            uniqueid = self.images_ids[idx]
            self.images[str(uniqueid)] = imgs[i]

    def __get_bounding_boxes_polygons(self, index):

        self.__retrieve_bounding_boxes(index)

        ret_poly = []

        uniqueid = self.images_ids[index]

        query = [{
            "FindImage": {
                "_ref": 1,
                "constraints": {
                    "_uniqueid": ["==", uniqueid]
                },
                "blobs": False,
            }
        }, {
            "FindBlob": {
                "is_connected_to": {
                    "ref": 1
                },
                "constraints": {
                    "type": ["==", "segmentation"]
                },
                "blobs": True,
            }
        }]

        res, polygons = self.db_connector.query(query)

        ret_poly.append(polygons)

        uniqueid_str = str(uniqueid)
        self.images_bboxes[uniqueid_str]["polygons"] = polygons

        return polygons

    def __display_segmentation(self, index):

        polygons = self.__get_bounding_boxes_polygons(index)

        image = self.get_image_by_index(index)

        img_stringIO = BytesIO(image)
        I = io.imread(img_stringIO)

        fig1, ax1 = plt.subplots()
        plt.imshow(I)
        plt.axis('off')

        for poly in polygons:

            sample = np.frombuffer(poly, dtype=np.float32)

            c = (np.random.random((1, 3)) * 0.6 + 0.4).tolist()[0]
            color = []
            color.append(c)

            polygon_points = []
            poly = np.array(sample).reshape((int(len(sample) / 2), 2))
            polygon_points.append(Polygon(poly))

            ax = plt.gca()
            ax.set_autoscale_on(False)

            p = PatchCollection(
                polygon_points, facecolor=color, linewidths=0, alpha=0.4)
            ax.add_collection(p)

            p = PatchCollection(polygon_points, facecolor='none',
                                edgecolors=color, linewidths=2)
            ax.add_collection(p)

    def __retrieve_bounding_boxes(self, index):

        self.images_bboxes = {}

        if index > len(self.images_ids):
            print("Index error when retrieving bounding boxes")
            return

        uniqueid = self.images_ids[index]

        query = [{
            "FindImage": {
                "_ref": 1,
                "constraints": {
                    "_uniqueid": ["==", uniqueid]
                },
                "blobs": False,
            }
        }, {
            "FindBoundingBox": {
                "image": 1,
                "_ref": 2,
                "blobs": False,
                "coordinates": True,
                "labels": True,
            }
        }]

        try:
            res, images = self.db_connector.query(query)

            bboxes = []
            tags   = []
            for bbox in res[1]["FindBoundingBox"]["entities"]:
                bboxes.append(bbox["_coordinates"])
                tags.append(bbox[self.bbox_label_prop])

            uniqueid_str = str(uniqueid)
            self.images_bboxes[uniqueid_str] = {}
            self.images_bboxes[uniqueid_str]["bboxes"] = bboxes
            self.images_bboxes[uniqueid_str]["tags"]   = tags

        except:
            print(self.db_connector.get_last_response_str())

    def total_results(self):

        return len(self.images_ids)

    def get_image_by_index(self, index):

        if index >= len(self.images_ids):
            print("Index is incorrect")
            return

        uniqueid = self.images_ids[index]

        # If image is not retrieved, go and retrieve the batch
        if not str(uniqueid) in self.images:
            self.__retrieve_batch(index)

        if self.images[str(uniqueid)] == None:
            print("Image was not retrieved")

        return self.images[str(uniqueid)]

    def get_np_image_by_index(self, index):

        image = self.get_image_by_index(index)
        # Just decode the image from buffer
        nparr = np.frombuffer(image, dtype=np.uint8)
        image = cv2.imdecode(nparr, cv2.IMREAD_COLOR)

        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

        return image

    def get_bboxes_by_index(self, index):

        if not self.images_bboxes:
            self.__retrieve_bounding_boxes(index)

        try:
            bboxes = self.images_bboxes[str(self.images_ids[index])]
        except:
            print("Cannot retrieve requested bboxes")

        return bboxes

    # A new search will throw away the results of any previous search

    def search(self, constraints=None, operations=None, format=None, limit=None):

        self.constraints = constraints
        self.operations  = operations
        self.format      = format
        self.limit       = limit

        self.images = {}
        self.images_ids = []
        self.images_bboxes = {}

        query = {"FindImage": {}}

        if constraints:
            query["FindImage"]["constraints"] = constraints.constraints

        if format:
            query["FindImage"]["as_format"] = format

        query["FindImage"]["results"] = {}

        if limit:
            query["FindImage"]["results"]["limit"] = limit

        query["FindImage"]["results"]["list"] = []
        query["FindImage"]["results"]["list"].append(self.img_id_prop)

        # Only retrieve images when needed
        query["FindImage"]["blobs"] = False

        response, images = self.db_connector.query([query])

        try:
            entities = response[0]["FindImage"]["entities"]

            for ent in entities:
                self.images_ids.append(ent[self.img_id_prop])
        except:
            print("Error with search")

        self.search_result = response

    def get_similar_images(self, set_name, n_neighbors):

        imgs_return = Images(self.db_connector)

        for uniqueid in self.images_ids:

            query = [{
                "FindImage": {
                    "_ref": 1,
                    "constraints": {
                        "_uniqueid":  ["==", uniqueid]
                    },
                    "blobs": False,
                }
            }, {
                "FindDescriptor": {
                    "set": set_name,
                    "is_connected_to": {
                        "ref": 1,
                    },
                    "blobs": True
                }
            }]

            response, blobs = self.db_connector.query(query)

            query = [{
                "FindDescriptor": {
                    "_ref": 1,
                    "set": set_name,
                    "k_neighbors": n_neighbors + 1,
                    "blobs":     False,
                    "distances": True,
                    "ids":       True,
                }
            }, {
                "FindImage": {
                    "is_connected_to": {
                        "ref": 1,
                    },
                    "results": {
                        "list": ["_uniqueid"]
                    }
                }
            }]

            response, blobs = self.db_connector.query(query, blobs)

            try:
                entities = response[1]["FindImage"]["entities"]

                for ent in entities[1:]:
                    imgs_return.images_ids.append(ent[self.img_id_prop])

            except:
                print("Error with similarity search")

        return imgs_return

    def display(self, show_bboxes=False, show_segmentation=False, limit=None):

        if not limit:
            limit = self.display_limit
            if self.display_limit < len(self.images_ids):
                print("Showing only first", limit, "results.")

        for i in range(len(self.images_ids)):

            if limit == 0:
                break
            limit -= 1

            uniqueid = str(self.images_ids[i])
            image    = self.get_image_by_index(i)

            # Just decode the image from buffer
            nparr = np.frombuffer(image, dtype=np.uint8)
            image = cv2.imdecode(nparr, cv2.IMREAD_COLOR)

            if show_bboxes:
                if not str(uniqueid) in self.images_bboxes:
                    self.__retrieve_bounding_boxes(i)

                bboxes = self.images_bboxes[uniqueid]["bboxes"]
                tags   = self.images_bboxes[uniqueid]["tags"]

                # image = cv2.resize(image, (0,0), fx=1.0,fy=1.0)
                # rgb   = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

                # Draw a rectangle around the faces
                counter = 0
                for coords in bboxes:
                    left   = coords["x"]
                    top    = coords["y"]
                    right  = coords["x"] + coords["width"]
                    bottom = coords["y"] + coords["height"]
                    cv2.rectangle(image, (left, top), (right, bottom),
                                  (0, 255, 0), 2)

                    y = top - 15 if top - 15 > 15 else top + 15

                    cv2.putText(image, tags[counter], (left, y),
                                cv2.FONT_HERSHEY_SIMPLEX, 0.75, (0, 255, 0), 2)
                    counter += 1

                image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

                fig1, ax1 = plt.subplots()
                plt.imshow(image), plt.axis("off")

            elif show_segmentation:
                self.__display_segmentation(i)

            else:

                image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

                fig1, ax1 = plt.subplots()
                plt.imshow(image), plt.axis("off")

    def get_props_names(self):

        dbutils = Utils.Utils(self.db_connector)
        schema = dbutils.get_schema()

        try:
            dictio = schema["entities"]["classes"]["_Image"]["properties"]
            props_array = [key for key, val in dictio.items()]
        except:
            props_array = []
            print("Cannot retrieve properties")

        return props_array

    def get_properties(self, prop_list=[]):

        if len(prop_list) == 0:
            return {}

        return_dictionary = {}

        try:
            for uniqueid in self.images_ids:

                query = [{
                    "FindImage": {
                        "_ref": 1,
                        "constraints": {
                            "_uniqueid": ["==", uniqueid]
                        },
                        "blobs": False,
                        "results": {
                            "list": prop_list
                        }
                    }
                }]

                res, images = self.db_connector.query(query)

                return_dictionary[str(
                    uniqueid)] = res[0]["FindImage"]["entities"][0]
        except:
            print("Cannot retrieved properties")

        return return_dictionary



import os
import math

import numpy as np
import cv2

from aperturedb import Images

import torch
from torch.utils import data
from torchvision import transforms

DEFAULT_BATCH_SIZE = 50


class ApertureDBDatasetConstraints(data.Dataset):

    # initialise function of class
    def __init__(self, db, constraints):

        self.imgs_handler = Images.Images(db)
        self.imgs_handler.search(constraints=constraints)

    def __getitem__(self, index):

        if index >= self.imgs_handler.total_results():
            raise StopIteration

        img = self.imgs_handler.get_np_image_by_index(index)

        # This is temporary until we define a good, generic way, of
        # retriving a label associated with the image.
        label = "none"
        return img, label

    def __len__(self):

        return self.imgs_handler.total_results()


class ApertureDBDataset(data.Dataset):

    # initialise function of class
    def __init__(self, db, query, label_prop=None):

        self.db = db.create_new_connection()
        self.query = query
        self.find_image_idx = None
        self.total_elements = 0
        self.batch_size     = DEFAULT_BATCH_SIZE
        self.batch_images   = []
        self.batch_start    = 0
        self.batch_end      = 0
        self.label_prop     = label_prop

        self.prev_requested   = -1
        self.sequence_counter = DEFAULT_BATCH_SIZE

        for i in range(len(query)):

            name = list(query[i].keys())[0]
            if name == "FindImage":
                self.find_image_idx = i

        if self.find_image_idx is None:
            print("Query error. The query must containt one FindImage command")
            raise Exception('Query Error')

        if not "results" in self.query[self.find_image_idx]["FindImage"]:
            self.query[self.find_image_idx]["FindImage"]["results"] = {}

        self.query[self.find_image_idx]["FindImage"]["batch"] = {}

        try:
            r, b = self.db.query(self.query)
            batch = r[self.find_image_idx]["FindImage"]["batch"]
            self.total_elements = batch["total_elements"]
        except:
            print("Query error:")
            print(self.query)
            print(self.db.get_last_response_str())
            raise

    def __getitem__(self, index):

        if index == self.prev_requested + 1:
            self.sequence_counter += 1
        else:
            self.sequence_counter = 0

        if self.sequence_counter >= DEFAULT_BATCH_SIZE:
            self.batch_size = DEFAULT_BATCH_SIZE
        else:
            self.batch_size = 1

        self.prev_requested = index

        if index >= self.total_elements:
            raise StopIteration

        if not self.is_in_range(index):
            self.get_batch(index)

        idx = index % self.batch_size
        img   = self.batch_images[idx]
        label = self.batch_labels[idx]

        nparr = np.frombuffer(img, dtype=np.uint8)
        img   = cv2.imdecode(nparr, cv2.IMREAD_COLOR)
        img   = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)

        return img, label

    def __len__(self):

        return self.total_elements

    def is_in_range(self, index):

        if index >= self.batch_start and index < self.batch_end:
            return True

        return False

    def get_batch(self, index):

        total_batches = math.ceil(self.total_elements / self.batch_size)
        batch_idx     = math.floor(index / self.batch_size)

        query  = self.query
        qbatch = query[self.find_image_idx]["FindImage"]["batch"]
        qbatch["batch_size"] = self.batch_size
        qbatch["batch_id"]   = batch_idx

        query[self.find_image_idx]["FindImage"]["batch"] = qbatch

        try:

            # This is to handle potential issues with
            # disconnection/timeout and SSL context on multiprocessing
            connection_ok = False
            try:
                r, b = self.db.query(query)
                connection_ok = True
            except:
                # Connection failed, we retry just once to re-connect
                self.db = self.db.create_new_connection()

            if not connection_ok:
                # Connection failed, we have reconnected, we try again.
                r, b = self.db.query(query)

            if len(b) == 0:
                print("index:", index)
                raise Exception("No results returned from ApertureDB")

            self.batch_images = b
            self.batch_start  = self.batch_size * batch_idx
            self.batch_end    = self.batch_start + len(b)

            if self.label_prop:
                entities = r[self.find_image_idx]["FindImage"]["entities"]
                self.batch_labels = [l[self.label_prop] for l in entities]
            else:
                self.batch_labels = ["none" for l in range(len(b))]
        except:
            print("Query error:")
            print(self.db.get_last_response_str())
            raise



import math
import time
from threading import Thread

from aperturedb import Parallelizer

import numpy as np


class ParallelQuery(Parallelizer.Parallelizer):

    """**Parallel and Batch Querier for ApertureDB**"""

    def __init__(self, db, dry_run=False):

        super().__init__()

        self.db = db.create_new_connection()

        self.dry_run = dry_run

        self.type = "query"

    def generate_batch(self, data):
        """
            Here we flatten the individual queries to run them as
            a single query in a batch
        """
        q     = [cmd for query in data for cmd in query[0]]
        blobs = [blob for query in data for blob in query[1]]

        return q, blobs

    def do_batch(self, db, data):

        q, blobs = self.generate_batch(data)

        if not self.dry_run:
            r, b = db.query(q, blobs)
            if not db.last_query_ok():
                self.error_counter += 1
            query_time = db.get_last_query_time()
        else:
            query_time = 1

        # append is thread-safe
        self.times_arr.append(query_time)

    def worker(self, thid, generator, start, end):

        # A new connection will be created for each thread
        db = self.db.create_new_connection()

        total_batches = (end - start) // self.batchsize

        if (end - start) % self.batchsize > 0:
            total_batches += 1

        for i in range(total_batches):

            batch_start = start + i * self.batchsize
            batch_end   = min(batch_start + self.batchsize, end)

            try:
                # This can be done with slices instead of list comprehension.
                # but required support from generator.
                data_for_query = [generator[idx]
                                  for idx in range(batch_start, batch_end)]
                self.do_batch(db, data_for_query)
            except Exception as e:
                print(e)
                self.error_counter += 1

            if thid == 0 and self.stats:
                self.pb.update((i + 1) / total_batches)

    def query(self, generator, batchsize=1, numthreads=4, stats=False):

        self.run(generator, batchsize, numthreads, stats)

    def print_stats(self):

        times = np.array(self.times_arr)
        total_queries_exec = len(times)

        print("============ ApertureDB Parallel Query Stats ============")
        print("Total time (s):", self.total_actions_time)
        print("Total queries executed:", total_queries_exec)

        if total_queries_exec == 0:
            print("All queries failed!")

        else:
            print("Avg Query time (s):", np.mean(times))
            print("Query time std:", np.std(times))
            print("Avg Query Throughput (q/s)):",
                  1 / np.mean(times) * self.numthreads)

            msg = "(" + self.type + "/s):"
            print("Overall throughput", msg,
                  self.total_actions / self.total_actions_time)

            if self.error_counter > 0:
                print("Total errors encountered:", self.error_counter)
                print("Errors (%):", 100 *
                      self.error_counter / total_queries_exec)

        print("=========================================================")



import numpy as np

from aperturedb import ParallelLoader
from aperturedb import CSVParser

HEADER_PATH  = "filename"
HEADER_INDEX = "index"
HEADER_SET   = "set"
HEADER_LABEL = "label"
PROPERTIES   = "properties"
CONSTRAINTS  = "constraints"


class DescriptorGeneratorCSV(CSVParser.CSVParser):
    """
        **ApertureDB Descriptor Data loader.**

    .. note::
        Expects a csv file with the following columns, and a numpy
        array file "npz" for the descriptors:

            ``filename``, ``index``, ``set``, ``label``, ``PROP_NAME_N``, ``constraint_PROP1``, ``connect_to``

    Example csv file::

        filename,index,set,label,isTable,connect_VD:ROI,connect_VD:ROI@index
        /mnt/data/embeddings/kitchen.npz,0,kitchen,kitchen_table,True,has_descriptor,0
        /mnt/data/embeddings/dining_chairs.npz,1,dining_chairs,special_chair,False,has_descriptor,1
        ...
    """

    def __init__(self, filename):

        super().__init__(filename)

        self.npy_arrays = {}
        self.has_label = False

        self.props_keys       = [x for x in self.header[3:]
                                 if not x.startswith(CSVParser.CONTRAINTS_PREFIX)]
        self.constraints_keys = [x for x in self.header[3:]
                                 if x.startswith(CSVParser.CONTRAINTS_PREFIX)]

    def __getitem__(self, idx):

        filename = self.df.loc[idx, HEADER_PATH]
        index    = self.df.loc[idx, HEADER_INDEX]
        desc_set = self.df.loc[idx, HEADER_SET]

        descriptor, desc_ok = self.load_descriptor(filename, index)

        if not desc_ok:
            print("Error loading descriptor: " + filename + ":" + index)
            raise Exception("Error loading descriptor: " +
                            filename + ":" + index)

        data = {
            "descriptor_blob": descriptor,
            "set": desc_set
        }

        if self.has_label:
            data[HEADER_LABEL] = self.df.loc[idx, HEADER_LABEL]

        properties  = self.parse_properties(self.df, idx)
        constraints = self.parse_constraints(self.df, idx)

        if properties:
            data[PROPERTIES] = properties

        if constraints:
            data[CONSTRAINTS] = constraints

        return data

    # This function can be re-defined by user when they have
    # app-specific structure on the npz file.
    def retrieve_by_index(self, filename, index):

        try:
            desc = self.npy_arrays[filename][index]
        except:
            err_msg = "Cannot retrieve descriptor {} from {}".format(
                str(index), filename)
            raise Exception(err_msg)

        return desc

    def load_descriptor(self, filename, index):

        if filename not in self.npy_arrays:
            self.npy_arrays[filename] = np.load(filename)

        # Can be defined by the user.
        descriptor = self.retrieve_by_index(filename, index)

        if len(descriptor) < 0:
            return [], False

        descriptor = descriptor.astype('float32').tobytes()

        return descriptor, True

    def validate(self):

        self.header = list(self.df.columns.values)

        if self.header[3] == HEADER_LABEL:
            self.has_label = True

        if self.header[0] != HEADER_PATH:
            raise Exception(
                "Error with CSV file field: filename. Must be first field")
        if self.header[1] != HEADER_INDEX:
            raise Exception(
                "Error with CSV file field: index. Must be second field")
        if self.header[2] != HEADER_SET:
            raise Exception(
                "Error with CSV file field: set. Must be third field")


class DescriptorLoader(ParallelLoader.ParallelLoader):
    """**ApertureDB Descriptor Loader.**

    This class is to be used in combination with a "generator".
    The generator must be an iterable object that generated "descriptor_data"
    elements.

    Example::

            image_data = {
                "label":       label,
                "properties":  properties,
                "constraints": constraints,
                "descriptor_blob": (bytes),
            }
    """

    def __init__(self, db, dry_run=False):

        super().__init__(db, dry_run=dry_run)

        self.type = "descriptor"

    def generate_batch(self, image_data):

        q = []
        blobs = []

        for data in image_data:

            ai = {
                "AddDescriptor": {
                    "set": data[HEADER_SET]
                }
            }

            if "properties" in data:
                ai["AddDescriptor"]["properties"] = data["properties"]
            if "constraints" in data:
                ai["AddDescriptor"]["if_not_found"] = data["constraints"]
            if "label" in data:
                ai["AddDescriptor"]["label"] = data["label"]

            if "descriptor_blob" not in data or len(data["descriptor_blob"]) == 0:
                print("WARNING: Skipping empty descriptor.")
                continue

            blobs.append(data["descriptor_blob"])
            q.append(ai)

        if self.dry_run:
            print(q)

        return q, blobs



import math
import time
from threading import Thread

import numpy as np

from aperturedb import ProgressBar


class Parallelizer:
    """**Generic Parallelizer**

    A parallelizer converts a series of operations to be executed and partitions it into
    bataches, to be execued by multiple threads of execution.

    .. image:: /_static/parallelizer.svg

    """

    def __init__(self, progress_to_file=""):

        # Default Values
        self.batchsize  = 1
        self.numthreads = 1

        self.total_actions = 0
        self.times_arr = []
        self.total_actions_time = 0
        self.error_counter  = 0

        if progress_to_file:
            self.pb = ProgressBar.ProgressBar(progress_to_file)
        else:
            self.pb = ProgressBar.ProgressBar()

    def get_times(self):

        return self.times_arr

    def run(self, generator, batchsize, numthreads, stats):

        self.times_arr  = []
        self.batchsize  = batchsize
        self.numthreads = numthreads
        self.stats      = stats
        self.total_actions = len(generator)

        start_time = time.time()

        if self.total_actions < batchsize:
            elements_per_thread = self.total_actions
            self.numthreads = 1
        else:
            elements_per_thread = math.ceil(
                self.total_actions / self.numthreads)

        thread_arr = []
        for i in range(self.numthreads):
            idx_start = i * elements_per_thread
            idx_end   = min(idx_start + elements_per_thread,
                            self.total_actions)

            thread_add = Thread(target=self.worker,
                                args=(i, generator, idx_start, idx_end))
            thread_arr.append(thread_add)

        a = [th.start() for th in thread_arr]
        a = [th.join() for th in thread_arr]

        self.total_actions_time = time.time() - start_time

        if self.stats:
            self.print_stats()

    def print_stats(self):
        """
            Must be implemented by child class
        """
        pass



from aperturedb import ParallelLoader
from aperturedb import CSVParser

HEADER_X_POS  = "x_pos"
HEADER_Y_POS  = "y_pos"
HEADER_WIDTH  = "width"
HEADER_HEIGHT = "height"
IMG_KEY_PROP  = "img_key_prop"
IMG_KEY_VAL   = "img_key_value"


class BBoxGeneratorCSV(CSVParser.CSVParser):
    """**ApertureDB BBox Data loader.**

    .. note::
        Expects a csv file with the following columns:

            ``IMG_KEY``, ``x_pos``, ``y_pos``, ``width``, ``height``, ``BBOX_PROP_NAME_1``, ... ``BBOX_PROP_NAME_N``

    IMG_KEY column has the property name of the image property that
    the bounding box will be connected to, and each row has the value
    that will be used for finding the image.

    x_pos,y_pos,width,height are the coordinates of the bounding boxes,
    as integers (unit is in pixels)

    BBOX_PROP_NAME_N is an arbitrary name of the property of the bounding
    box, and each row has the value for that property.

    Example csv file::

        img_unique_id,x_pos,y_pos,width,height,type
        d5b25253-9c1e,257,154,84,125,manual
        d5b25253-9c1e,7,537,522,282,manual
        ...
    """

    def __init__(self, filename):

        super().__init__(filename)

        self.props_keys       = [x for x in self.header[5:]
                                 if not x.startswith(CSVParser.CONTRAINTS_PREFIX)]
        self.constraints_keys = [x for x in self.header[5:]
                                 if x.startswith(CSVParser.CONTRAINTS_PREFIX)]

        self.img_key = self.header[0]

    def __getitem__(self, idx):

        data = {
            "x":      int(self.df.loc[idx, HEADER_X_POS]),
            "y":      int(self.df.loc[idx, HEADER_Y_POS]),
            "width":  int(self.df.loc[idx, HEADER_WIDTH]),
            "height": int(self.df.loc[idx, HEADER_HEIGHT]),
        }

        val = self.df.loc[idx, self.img_key]

        if val != "":
            data[IMG_KEY_PROP] = self.img_key
            data[IMG_KEY_VAL]  = val

        properties  = self.parse_properties(self.df, idx)
        constraints = self.parse_constraints(self.df, idx)

        if properties:
            data[CSVParser.PROPERTIES] = properties

        if constraints:
            data[CSVParser.CONSTRAINTS] = constraints

        return data

    def validate(self):

        self.header = list(self.df.columns.values)

        if self.header[1] != HEADER_X_POS:
            raise Exception("Error with CSV file field: " + HEADER_X_POS)
        if self.header[2] != HEADER_Y_POS:
            raise Exception("Error with CSV file field: " + HEADER_Y_POS)
        if self.header[3] != HEADER_WIDTH:
            raise Exception("Error with CSV file field: " + HEADER_WIDTH)
        if self.header[4] != HEADER_HEIGHT:
            raise Exception("Error with CSV file field: " + HEADER_HEIGHT)


class BBoxLoader(ParallelLoader.ParallelLoader):

    def __init__(self, db, dry_run=False):

        super().__init__(db, dry_run=dry_run)

        self.type = "bbox"

    def generate_batch(self, bbox_data):

        q = []

        ref_counter = 1
        for data in bbox_data:

            # TODO we could reuse image references within the batch
            # instead of creating a new find for every image.
            img_ref = ref_counter
            ref_counter += 1
            fi = {
                "FindImage": {
                    "_ref": img_ref,
                }
            }

            if IMG_KEY_PROP in data:
                key = data[IMG_KEY_PROP]
                val = data[IMG_KEY_VAL]
                constraints = {}
                constraints[key] = ["==", val]
                fi["FindImage"]["constraints"] = constraints

            q.append(fi)

            ai = {
                "AddBoundingBox": {
                    "image": img_ref,
                    "rectangle": {
                        "x":      data["x"],
                        "y":      data["y"],
                        "width":  data["width"],
                        "height": data["height"],
                    },
                }
            }

            if "properties" in data:
                props = data[CSVParser.PROPERTIES]
                if "_label" in props:
                    ai["AddBoundingBox"]["label"] = props["_label"]
                    props.pop("_label")

                # Check if props is not empty after removing "_label"
                if props:
                    ai["AddBoundingBox"]["properties"] = props

            q.append(ai)

        if self.dry_run:
            print(q)

        return q, []



import time
import unittest

import dbinfo

from aperturedb import Connector


class TestBase(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # ApertureDB Server Info
        self.db_host  = dbinfo.DB_HOST
        self.db_port  = dbinfo.DB_PORT
        self.user     = dbinfo.DB_USER
        self.password = dbinfo.DB_PASSWORD

        # Config params
        self.batchsize  = 99
        self.numthreads = 31
        self.stats      = False

        db_up = False
        attempts = 0
        while(not db_up):
            try:
                db = Connector.Connector(self.db_host, self.db_port,
                                         user=self.user, password=self.password)
                db_up = True
                if (attempts > 0):
                    print("Connection to ApertureDB successful.")
            except:
                print("Attempt", attempts,
                      "to connect to ApertureDB failed, retrying...")
                # db.print_last_response()

                attempts += 1
                time.sleep(1)  # sleeps 1 second

            if attempts > 10:
                print("Failed to connect to ApertureDB after 10 attempts")
                exit()

    def create_connection(self):
        return Connector.Connector(self.db_host, self.db_port, self.user, self.password)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="aperturedb",
    version="0.1.7",
    description="ApertureDB Client Module",
    install_requires=['vdms', 'scikit-image', 'image', 'requests', 'boto3',
                      'opencv-python', 'numpy', 'matplotlib', 'pandas'],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/aperture-data/aperturedb-python",
    license="Apache",
    packages=setuptools.find_packages(),
    python_requires='>=2.6, !=3.0.*, !=3.1.*, !=3.2.*, <4',
    author="Luis Remis",
    author_email="luis@aperturedata.io",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
)
